""" Pydantic and Sqlalchemy models for the database

7. PV system for storing PV data (pv.py)
8. PV yield for storing PV data (pv.py)

"""
import logging
from datetime import datetime
from typing import Optional

from pydantic import Field, validator
from sqlalchemy import Column, DateTime, Float, ForeignKey, Index, Integer, String
from sqlalchemy.orm import relationship

from nowcasting_datamodel.models.base import Base_PV
from nowcasting_datamodel.models.utils import CreatedMixin, EnhancedBaseModel
from nowcasting_datamodel.utils import datetime_must_have_timezone

logger = logging.getLogger(__name__)

providers = ["pvoutput.org"]

########
# 7. PV Metadata
########


class PVSystemSQL(Base_PV, CreatedMixin):
    """Metadata for PV data"""

    __tablename__ = "pv_system"

    id = Column(Integer, primary_key=True)
    pv_system_id = Column(Integer, index=True)
    provider = Column(String)
    latitude = Column(Float)
    longitude = Column(Float)
    name = Column(String, nullable=True)
    orientation = Column(Float, nullable=True)
    status_interval_minutes = Column(Integer, nullable=True)
    installed_capacity_kw = Column(
        Float,
        nullable=True,
    )

    pv_yield = relationship("PVYieldSQL", back_populates="pv_system")


class PVSystem(EnhancedBaseModel):
    """Metadata for PV data"""

    pv_system_id: int = Field(..., description="The PV system id")
    provider: str = Field(..., description="The provider of the PV system")
    latitude: float = Field(None, description="The latitude of the PV system")
    longitude: float = Field(None, description="The longitude of the PV system")
    name: Optional[str] = Field(None, description="The PV system name")
    orientation: Optional[float] = Field(None, description="The orientation of the PV system")
    status_interval_minutes: Optional[float] = Field(
        None, description="The number of minutes for the pv data to be refreshed"
    )
    installed_capacity_kw: Optional[float] = Field(
        None, description="The capacity of the pv system in kw."
    )

    @validator("provider")
    def validate_provider(cls, v):
        """Validate the provider"""
        if v not in providers:
            raise Exception(f"Provider ({v}) must be in {providers}")
        return v

    def to_orm(self) -> PVSystemSQL:
        """Change model to PVSystemSQL"""
        return PVSystemSQL(
            pv_system_id=self.pv_system_id,
            provider=self.provider,
            latitude=self.latitude,
            longitude=self.longitude,
            name=self.name,
            orientation=self.orientation,
            status_interval_minutes=self.status_interval_minutes,
        )


########
# 8. PV Yield
########


class PVYieldSQL(Base_PV, CreatedMixin):
    """PV Yield data"""

    __tablename__ = "pv_yield"

    id = Column(Integer, primary_key=True)
    datetime_utc = Column(DateTime, index=True)
    solar_generation_kw = Column(String)

    # many (forecasts) to one (location)
    pv_system = relationship("PVSystemSQL", back_populates="pv_yield")
    pv_system_id = Column(Integer, ForeignKey("pv_system.id"), index=True)

    Index("ix_datetime_utc", datetime_utc.desc())


class PVYield(EnhancedBaseModel):
    """PV Yield data"""

    datetime_utc: datetime = Field(..., description="The timestamp of the pv system")
    solar_generation_kw: float = Field(..., description="The provider of the PV system")

    _normalize_target_time = validator("datetime_utc", allow_reuse=True)(
        datetime_must_have_timezone
    )

    pv_system: Optional[PVSystem] = Field(
        None,
        description="The PV system associated with this model",
    )

    @validator("solar_generation_kw")
    def validate_solar_generation_kw(cls, v):
        """Validate the solar_generation_kw field"""
        if v < 0:
            logger.debug(f"Changing solar_generation_kw ({v}) to 0")
            v = 0
        return v

    def to_orm(self) -> PVYieldSQL:
        """Change model to PVYieldSQL"""
        return PVYieldSQL(
            datetime_utc=self.datetime_utc,
            solar_generation_kw=self.solar_generation_kw,
        )



""" Utils functions for models """
import logging
from datetime import datetime, timezone

logger = logging.getLogger(__name__)


def datetime_must_have_timezone(cls, v: datetime):
    """Enforce that this variable must have a timezone"""
    if v.tzinfo is None:
        logger.debug(f"{v} must have a timezone, for cls {cls}." f"So we are going to add UTC ")
        v = v.replace(tzinfo=timezone.utc)
    return v


def convert_to_camelcase(snake_str: str) -> str:
    """Converts a given snake_case string into camelCase"""
    first, *others = snake_str.split("_")
    return "".join([first.lower(), *map(str.title, others)])



""" Method to update latest forecast values """
import logging
from datetime import datetime, timezone
from typing import List

from sqlalchemy import inspect
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm.session import Session

from nowcasting_datamodel.models.models import ForecastSQL, ForecastValueLatestSQL
from nowcasting_datamodel.read.read import get_latest_forecast, get_model

logger = logging.getLogger(__name__)


def get_historic_forecast(session: Session, forecast: ForecastSQL) -> ForecastSQL:
    """
    Get historic forecast

    :param session:
    :param gsp_id:
    :return:
    """

    gsp_id = forecast.location.gsp_id

    forecast_historic = get_latest_forecast(session=session, gsp_id=gsp_id, historic=True)

    if forecast_historic is None:
        logger.debug("Could not find a historic forecast, so will make one")

        forecast_historic = ForecastSQL(
            historic=True,
            forecast_creation_time=datetime.now(timezone.utc),
            location=forecast.location,
            input_data_last_updated=forecast.input_data_last_updated,
            model=get_model(session=session, name="historic", version="all"),
        )
    else:
        logger.debug("Found historic forecast")

    return forecast_historic


def upsert(session: Session, model, rows: List[dict]):
    """

    Upsert rows into model

    Insert or Update rows into model.
    This functions checks the primary keys, and if duplicate updates them.

    :param session: sqlalchemy Session
    :param model: the model
    :param rows: the rows we are going to update
    """
    table = model.__table__
    stmt = insert(table)
    primary_keys = [key.name for key in inspect(table).primary_key]
    update_dict = {c.name: c for c in stmt.excluded if not c.primary_key}

    if not update_dict:
        raise ValueError("insert_or_update resulted in an empty update_dict")

    stmt = stmt.on_conflict_do_update(index_elements=primary_keys, set_=update_dict)
    session.execute(stmt, rows)


def update_forecast_latest(forecast: ForecastSQL, session: Session):
    """
    Update the forecast_values table

    1. Get the Forecast object that is for the latest,
        and don't load children
    2. create ForecastValueLatestSQL objects
    3. upsert them (update and or insert)

    :param forecast:
    :return:
    """

    # 1. get forecast object
    forecast_historic = get_historic_forecast(session=session, forecast=forecast)

    # 2. create forecast value latest
    forecast_values_dict = []
    for forecast_value in forecast.forecast_values:

        forecast_value_dict = {}
        for v in ["target_time", "expected_power_generation_megawatts"]:
            forecast_value_dict[v] = getattr(forecast_value, v)

        forecast_value_dict["gsp_id"] = forecast.location.gsp_id
        forecast_value_dict["forecast_id"] = forecast_historic.id

        forecast_values_dict.append(ForecastValueLatestSQL(**forecast_value_dict).__dict__)

    # upsert forecast values
    upsert(session=session, model=ForecastValueLatestSQL, rows=forecast_values_dict)

    # update forecast creation time
    forecast_historic.forecast_creation_time = datetime.now(tz=timezone.utc)
    session.commit()


def update_all_forecast_latest(forecasts: List[ForecastSQL], session: Session):
    """
    Update all latest forecasts
    """

    for forecast in forecasts:
        update_forecast_latest(forecast=forecast, session=session)
